#include "voxmodel/activeindex.h"

template <class T> 
ActiveIndex< T >::~ActiveIndex()
{
  delete [] data;
  delete [] unused;
}
 
template <class T>
ActiveIndex< T >::ActiveIndex (unsigned initial_size, unsigned inc_size)
{
  unsigned i;
  data = new T[initial_size];
  unused = new unsigned[initial_size];
  if ((data==0)||(unused==0)) {
    cout<<"--- "<<T::name()<<" information ---\n";
    cout<<"Out of memory!\n";
    exit(-1);
  }
  for (i=0;i<initial_size;i++) unused[initial_size-1-i]=i;
  cout<<"--- "<<T::name()<<" information ---\n";
  cout<<"Initial memory size is : "<<initial_size<<endl;
  available=initial_size;
  size_increment=inc_size;
  current_size=initial_size;
  max_accessed=0;
  n_fixed=0;
}

template <class T>
void ActiveIndex< T >::expand ()
{
  T *swap;
  unsigned *swap2, i;
  
  cout << "Expanding memory allocation ("<<T::name()<<") ...";
  swap = new T[current_size+size_increment];
  swap2 = new unsigned[current_size+size_increment];

  if ((swap==0)||(swap2==0)) {
    cout<<"Out of memory!\n";
    exit(-1);
  }

  for (i=0;i<current_size;i++) swap[i]=data[i]; // = operator must be defined for class T

  delete [] data;
  delete [] unused;
  data=swap;
  unused=swap2;

  for (i=0;i<size_increment;i++) unused[size_increment-1-i]=current_size++;
  available=size_increment;
  cout<<"done\n";
  info();
}

template <class T>
void ActiveIndex< T >::edit (unsigned index, T &new_data) { 
  if (index<n_fixed) cout<<"Attempt to edit fixed entry! ("<<T::name()<<")\n";
  else {
    if (index<current_size) data[index]=new_data; // = operator must be defined for class T
    else {
      cout<<"ActiveIndex error: Attempt to edit non existing data entry ("<<T::name()<<")\n";
      exit(-1);
    }
  }
}

template <class T>
unsigned ActiveIndex< T >::push (T &new_data) {
  if (available==0) expand();
  unsigned i=unused[--available];
  data[i]=new_data;             // = operator must be defined for class T
  if ((i+1)>max_accessed) max_accessed=i+1;
  return i;
} 

template <class T>
void ActiveIndex< T >::info ()
{
  cout<<"--- "<<T::name()<<" information ---\n";
  // T must have static name() function to return its class name
  cout<<"Data cells allocated             : "<<current_size<<endl;
  cout<<"Max. data cell used              : "<<max_accessed<<endl;
  cout<<"Current number of available cells: "<<available<<endl;
  cout<<"Next available slot              : "<<unused[available-1]<<endl;
}
